#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Implements variable dimension sizes for annotations. """
from typing import Any
from sympy import symbols


def __getattr__(name: str) -> Any:
    """ Yields the dims. """
    return symbols(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" A module for storing python objects for internal asta use. """
from typing import Dict, Any

# pylint: disable=redefined-outer-name

storage: Dict[str, Any]


def __getattr__(name: str) -> Any:
    """ Yields the objects. """
    try:
        try:
            return storage[name]
        except KeyError:
            return None
    except NameError:
        return None


def __setattr__(name: str, value: Any) -> None:
    """ Sets the objects. """
    try:
        storage
    except NameError:
        storage = {}
    storage[name] = value



""" Functions for generating typechecker output. """
import inspect
from typing import Any, Dict, Set, List, FrozenSet, Union

import numpy as np
from sympy.core.expr import Expr
from sympy.core.symbol import Symbol

from oxentiel import Oxentiel

from asta.array import Array
from asta.classes import SubscriptableMeta
from asta.constants import torch, tf, Color, _TORCH_IMPORTED, _TENSORFLOW_IMPORTED


if _TORCH_IMPORTED:
    from asta.tensor import Tensor

if _TENSORFLOW_IMPORTED:
    from asta.tftensor import TFTensor

FAIL = f"{Color.RED}FAILED{Color.END}"
PASS = f"{Color.GREEN}PASSED{Color.END}"


def get_type_name(type_: type) -> str:
    """ ``typing.*`` types don't have a __name__ on Python 3.7+. """
    # pylint: disable=protected-access
    name: str = getattr(type_, "__name__", None)
    if name is None:
        name = type_._name
    return name


def qualified_name(obj: Any) -> str:
    """
    Return the qualified name (e.g. package.module.Type) for the given object.

    Builtins and types from the :mod:`typing` package get special treatment by having the module
    name stripped from the generated name.

    """
    type_ = obj if inspect.isclass(obj) else type(obj)
    module = type_.__module__
    qualname: str = type_.__qualname__
    return qualname if module in ("typing", "builtins") else f"{module}.{qualname}"


def type_representation(arg: Any) -> str:
    """ Get a string representation of an argument including dtype and shape. """
    rep = repr(type(arg))
    if hasattr(arg, "shape"):
        shape = arg.shape
    if hasattr(arg, "dtype"):
        dtype = arg.dtype
    if isinstance(arg, np.ndarray):
        astatype = Array[dtype, shape]  # type: ignore[valid-type, type-arg]
        rep = repr(astatype).replace("asta", "numpy")
    elif isinstance(arg, torch.Tensor):
        astatype = Tensor[dtype, shape]  # type: ignore
        rep = repr(astatype).replace("asta", "torch")
    elif isinstance(arg, tf.Tensor):
        astatype = TFTensor[dtype, shape]  # type: ignore
        rep = repr(astatype).replace("asta", "tf")
    return rep


def pass_argument(name: str, ann: SubscriptableMeta, rep: str, ox: Oxentiel) -> None:
    """ Print typecheck pass notification for arguments. """
    msg = f"{PASS}: Argument '{name}' matched parameter '{ann}' "
    msg += f"with actual type: '{rep}'"
    handle_pass(msg, ox)


def pass_return(name: str, ann: SubscriptableMeta, rep: str, ox: Oxentiel) -> None:
    """ Print typecheck pass notification for return values. """
    msg = f"{PASS}: Return {name} matched return type '{ann}' with actual type: '{rep}'"
    handle_pass(msg, ox)


def fail_argument(name: str, ann: SubscriptableMeta, rep: str, ox: Oxentiel) -> None:
    """ Print/raise typecheck fail error for arguments. """
    err = f"{FAIL}: Argument '{name}' has wrong type. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_return(name: str, ann: SubscriptableMeta, rep: str, ox: Oxentiel) -> None:
    """ Print/raise typecheck fail error for return values. """
    err = f"{FAIL}: Return {name} has wrong type. Expected type: "
    err += f"'{ann}' Actual type: '{rep}'"
    handle_error(err, ox)


def fail_uninitialized(name: str, ox: Oxentiel) -> None:
    """ Print/raise typecheck fail error for uninitialized placeholder. """
    err = f"{FAIL}: Uninitialized placeholder '{name}'"
    handle_error(err, ox)


def fail_numerical_expression(
    item: str, expression: Union[Expr, Symbol], ox: Oxentiel
) -> None:
    """ For when refreshed expression yields non-integer ``Number``. """
    err = f"{FAIL}: Refreshed value '{expression}' for expression '{item}' "
    err += f"does not yield an integer."
    handle_error(err, ox)


def fail_tuple(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument is not a tuple. """
    err = f"{FAIL}: Argument '{name}' must be a Tuple. Actual type: '{rep}'"
    handle_error(err, ox)


def fail_empty_tuple(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument is not an empty tuple. """
    err = f"{FAIL}: Argument '{name}' must be an empty tuple. Actual type: '{rep}'"
    handle_error(err, ox)


def fail_tuple_length(
    name: str, length: int, ann_rep: str, ann_length: int, ox: Oxentiel
) -> None:
    """ Print/raise error when argument tuple has mismatched length. """
    err = f"{FAIL}: Tuple argument '{name}' has mismatched length ({length}). "
    err += f"Expected type: '{ann_rep}' with length {ann_length}"
    handle_error(err, ox)


def fail_namedtuple(name: str, ann_rep: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when NamedTuple fails isinstance check. """
    err = f"{FAIL}: Argument '{name}' has wrong type. Expected type: '{ann_rep}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_list(name: str, ann: Any, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when List fails isinstance check. """
    err = f"{FAIL}: Argument '{name}' must be a list. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_sequence(name: str, ann: Any, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when ``collections.abc.Sequence`` fails isinstance check. """
    err = f"{FAIL}: Argument '{name}' must be a sequence. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_dict(name: str, ann: Any, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when ``Dict[*]`` fails isinstance check. """
    err = f"{FAIL}: Argument '{name}' must be a dictionary. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_set(name: str, ann: Any, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when ``Set[*]`` fails isinstance check. """
    err = f"{FAIL}: Argument '{name}' must be a set. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_callable(name: str, ann: Any, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when ``Callable[*]`` argument is not callable. """
    err = f"{FAIL}: Argument '{name}' is not callable. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_too_many_args(
    name: str, expected_num_args: int, declared_num_args: int, ox: Oxentiel
) -> None:
    """ Print/raise error when ``Callable[*]`` argument has too few arguments. """
    err = f"{FAIL}: Callable '{name}' has too many arguments in its declaration. "
    err += f"Exepected: {expected_num_args} Declared: {declared_num_args}"
    handle_error(err, ox)


def fail_too_few_args(
    name: str, expected_num_args: int, declared_num_args: int, ox: Oxentiel
) -> None:
    """ Print/raise error when ``Callable[*]`` argument has too few arguments. """
    err = f"{FAIL}: Callable '{name}' has too few arguments in its declaration. "
    err += f"Exepected: {expected_num_args} Declared: {declared_num_args}"
    handle_error(err, ox)


def fail_protocol(name: str, ann_rep: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when ``Protocol[*]`` fails issubclass check. """
    err = f"{FAIL}: Argument '{name}' of type {rep} is not compatible "
    err += f"with the {ann_rep} protocol."
    handle_error(err, ox)


def fail_union(name: str, typelist: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when ``Union[*]`` fails typecheck. """
    err = f"{FAIL}: Argument '{name}' must be one of 'Union[{typelist}]'. "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_class(name: str, ann: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be a class but is not of type type. """
    err = f"{FAIL}: Argument '{name}' must be a class. Expected type: '{ann}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_subclass(name: str, supercls: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be subclass of ``expected_class``. """
    err = f"{FAIL}: Argument '{name}' should be a subclass of '{supercls}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_complex(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be a complex number. """
    err = f"{FAIL}: Argument '{name}' should be a complex number. "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_float(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be a floating point number. """
    err = f"{FAIL}: Argument '{name}' should be a float. "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_text_io(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be a text-based I/O object. """
    err = f"{FAIL}: Argument '{name}' should be a text-based I/O object. "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_binary_io(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be a binary I/O object. """
    err = f"{FAIL}: Argument '{name}' should be a binary I/O object. "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_io(name: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when argument should be an I/O object. """
    err = f"{FAIL}: Argument '{name}' should be an I/O object. "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def fail_literal(name: str, options: List[Any], arg: Any, ox: Oxentiel) -> None:
    """ Print/raise error when argument should match a given literal. """
    err = f"{FAIL}: Argument '{name}' should be one of '{options}' "
    err += f"Actual value: '{arg}'"
    handle_error(err, ox)


def fail_keys(
    name: str,
    expected_keys: FrozenSet[Any],
    existing_keys: FrozenSet[Any],
    ox: Oxentiel,
) -> None:
    """ Print/raise error when typed dict argument has extra or missing keys. """
    err = f"{FAIL}: Argument '{name}' should be a typed dictionary with keys: "
    err += f"'{expected_keys}'. Actual keys: '{existing_keys}'"
    handle_error(err, ox)


def fail_system(
    equations: Set[Expr],
    symbols: Set[Symbol],
    solutions: List[Dict[Symbol, int]],
    ox: Oxentiel,
) -> None:
    """ Print/raise typecheck fail error for uninitialized placeholder. """
    solution_err = f"No unique solution (found {len(solutions)})"
    err = f"{FAIL}: {solution_err} for system "
    err += f"'{equations}' of symbols '{symbols}'"
    handle_error(err, ox)


def fail_fallback(name: str, annrep: str, rep: str, ox: Oxentiel) -> None:
    """ Print/raise error when arbitrary value fails isinstance check. """
    err = f"{FAIL}: Argument '{name}' is not an instance of: '{annrep}' "
    err += f"Actual type: '{rep}'"
    handle_error(err, ox)


def handle_error(err: str, ox: Oxentiel) -> None:
    """ Either print or raise ``err``. """
    if not ox.print_passes and hasattr(ox, "decorated"):
        header = get_header(ox.decorated)
        print(header)
    if ox.raise_errors:
        raise TypeError(err)
    print(err)


def handle_pass(msg: str, ox: Oxentiel) -> None:
    """ Either print or raise ``err``. """
    if ox.print_passes:
        print(msg)


def get_header(decorated) -> str:  # type: ignore[no-untyped-def]
    """ Print the typecheck header. """
    core = f"asta::{decorated.__module__}.{decorated.__qualname__}()"
    bold_core = f"<{Color.BOLD}{core}{Color.END}>"
    min_pad_size = 10
    pad_size = 100 - (len(core) + 2)
    side_size = max(pad_size // 2, min_pad_size)
    pad_parity = pad_size % 2 if side_size > 10 else 0
    left_padding = "=" * side_size
    right_padding = "=" * (side_size + pad_parity)
    header = f"{left_padding}{bold_core}{right_padding}"
    return header



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" This module contains meta functionality for the ``Array`` type. """
import datetime
from abc import abstractmethod
from typing import List, Optional, Any, Tuple, Dict, Union

import numpy as np

from asta.utils import shapecheck, attrcheck
from asta.parser import parse_subscript
from asta.classes import SubscriptableMeta, GenericMeta
from asta.constants import (
    EllipsisType,
    NUMPY_DIM_TYPES,
    NP_UNSIZED_TYPE_KINDS,
)

# pylint: disable=unidiomatic-typecheck, too-few-public-methods, too-many-nested-blocks


class _ArrayMeta(SubscriptableMeta):
    """ A meta class for the ``Array`` class. """

    kind: str
    shape: tuple
    dtype: np.dtype
    kwattrs: Dict[str, Any]

    @classmethod
    @abstractmethod
    def _after_subscription(cls, item: Any) -> None:
        """ Method signature for subscript argument processing. """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_dtype(item: Any) -> Tuple[Optional[np.dtype], str]:
        """ Computes dtype. """
        raise NotImplementedError

    def __getitem__(cls, item: Any) -> GenericMeta:
        """ Defer to superclass, which calls ``cls._after_subscription()``. """
        return SubscriptableMeta.__getitem__(cls, item)

    def __eq__(cls, other: Any) -> bool:
        """ If the dtypes and shapes match, they should be equal. """
        if not isinstance(other, _ArrayMeta):
            return False
        if cls.shape != other.shape or cls.dtype != other.dtype:
            return False
        return True

    def __instancecheck__(cls, inst: Any) -> bool:
        """ Support expected behavior for ``isinstance(<array>, Array[<args>])``. """
        assert hasattr(cls, "kind")
        assert hasattr(cls, "shape")
        assert hasattr(cls, "dtype")
        assert hasattr(cls, "kwattrs")
        match = False
        if isinstance(inst, np.ndarray):
            match = True  # In case of an empty array or no ``cls.kind``.
            if inst.dtype.names:
                match = False

            if cls.kind and cls.kind != inst.dtype.kind:
                match = False

            # If we have ``cls.dtype``, we can be maximally precise.
            elif cls.dtype and cls.dtype != inst.dtype and cls.kind == "":
                match = False

            # Handle ellipses.
            else:
                shape_match, _ = shapecheck(inst.shape, cls.shape)
                attr_match, _ = attrcheck(inst, cls.kwattrs)
                match = shape_match and attr_match

        return match


class _Array(metaclass=_ArrayMeta):
    """ This class exists to keep the Array class as clean as possible. """

    NAME: str = "Array"
    DIM_TYPES: List[type] = NUMPY_DIM_TYPES
    _UNSIZED_TYPE_KINDS: Dict[type, str] = NP_UNSIZED_TYPE_KINDS

    kind: str = ""
    dtype: Optional[np.dtype] = None
    shape: Optional[Tuple] = None
    kwattrs: Optional[Dict[str, Any]] = None

    def __new__(cls, *args: Tuple[Any], **kwargs: Dict[str, Any]) -> Any:
        raise TypeError("Cannot instantiate abstract class 'Array'.")

    @staticmethod
    def get_dtype(item: Any) -> Tuple[Optional[np.dtype], str]:
        """ Computes dtype. """
        dtype = None
        kind = ""

        # Case where ``item`` is a dtype (``Array[np.dtype("complex128")]``).
        if isinstance(item, np.dtype):
            dtype = item

        # Case where ``item`` is a python3 type (``Array[int]``).
        elif isinstance(item, type):
            if item in NP_UNSIZED_TYPE_KINDS:
                generic_type = item
                kind = NP_UNSIZED_TYPE_KINDS[item]
            elif item == datetime.datetime:
                generic_type = np.datetime64
            elif item == datetime.timedelta:
                generic_type = np.timedelta64
            else:
                generic_type = item
            dtype = np.dtype(generic_type)

        return dtype, kind

    @classmethod
    def _after_subscription(
        cls, item: Union[type, Optional[Union[int, EllipsisType]]]  # type: ignore
    ) -> None:
        """ Set class attributes based on the passed dtype/dim data. """
        dtype, shape, kwattrs, kind = parse_subscript(cls, item, np.dtype)
        cls.dtype = dtype
        cls.shape = shape
        cls.kwattrs = kwattrs
        cls.kind = kind



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Configuration parsing and loading (credit to pylint authors). """
import os
import io
import collections
import configparser
from typing import Generator, Optional, Dict, Any

import toml
from oxentiel import Oxentiel

from asta import _internal


def _toml_has_config(path: str) -> bool:
    with open(path, "r") as toml_handle:
        content = toml.load(toml_handle)
        try:
            content["tool"]["asta"]
        except KeyError:
            return False

    return True


def _cfg_has_config(path: str) -> bool:
    parser = configparser.ConfigParser()
    parser.read(path)
    return any(section.startswith("asta.") for section in parser.sections())


def find_default_config_files() -> Generator[str, None, None]:
    """ Find all possible config files. """
    rc_names = ("astarc", ".astarc")
    config_names = rc_names + ("pyproject.toml", "setup.cfg")
    for config_name in config_names:
        if os.path.isfile(config_name):
            if config_name.endswith(".toml") and not _toml_has_config(config_name):
                continue
            if config_name.endswith(".cfg") and not _cfg_has_config(config_name):
                continue

            yield os.path.abspath(config_name)

    if os.path.isfile("__init__.py"):
        curdir = os.path.abspath(os.getcwd())
        while os.path.isfile(os.path.join(curdir, "__init__.py")):
            curdir = os.path.abspath(os.path.join(curdir, ".."))
            for rc_name in rc_names:
                rc_path = os.path.join(curdir, rc_name)
                if os.path.isfile(rc_path):
                    yield rc_path

    if "ASTARC" in os.environ and os.path.exists(os.environ["ASTARC"]):
        if os.path.isfile(os.environ["ASTARC"]):
            yield os.environ["ASTARC"]
    else:
        user_home = os.path.expanduser("~")
        if user_home not in ("~", "/root"):
            home_rc = os.path.join(user_home, ".astarc")
            if os.path.isfile(home_rc):
                yield home_rc
            home_rc = os.path.join(user_home, ".config", "astarc")
            if os.path.isfile(home_rc):
                yield home_rc

    if os.path.isfile("/etc/astarc"):
        yield "/etc/astarc"


def find_astarc() -> Optional[str]:
    """ Look for an ``astarc`` file. Return it if found, else return ``None``. """
    for config_file in find_default_config_files():
        if config_file.endswith("astarc"):
            return config_file
    return None


# pylint: disable=protected-access
def read_config(config_path: str) -> configparser.ConfigParser:
    """ Parse a config file from ``config_path``, given a file exists there. """
    parser = configparser.ConfigParser()
    if config_path.endswith(".toml"):
        with open(config_path, "r") as toml_file:
            content = toml.load(toml_file)

        try:
            sections_values = content["tool"]["asta"]
        except KeyError:
            pass
        else:
            for section, values in sections_values.items():
                parser._sections[section.upper()] = values  # type: ignore[attr-defined]
    else:
        # Use this encoding in order to strip the BOM marker, if any.
        with io.open(config_path, "r", encoding="utf_8_sig") as config_file:
            parser.read_file(config_file)

        # Normalize sections' titles.
        for sect, values in list(parser._sections.items()):  # type: ignore
            if sect.startswith("asta."):
                sect = sect[len("asta.") :]
            if not sect.isupper() and values:
                parser._sections[sect.upper()] = values  # type: ignore[attr-defined]
    return parser


def as_dict(config: configparser.ConfigParser) -> Dict[str, Dict[str, str]]:
    """ Returns a ``ConfigParser`` object as a dictionary. """
    dictionary: Dict[str, Dict[str, str]] = collections.OrderedDict()
    for section in config.sections():
        dictionary[section] = {}
        for key, val in config.items(section):
            dictionary[section][key] = val
    return dictionary


def get_ox() -> Oxentiel:
    """ Returns a configuration file. """
    ox = _internal.ox
    if not ox:
        path = find_astarc()
        parse = path and os.path.exists(path)
        if parse and path:
            parser = configparser.ConfigParser()
            parser = read_config(path)
            if "MASTER" in parser:
                settings = as_dict(parser)["MASTER"]
            else:
                settings = collections.OrderedDict()
        else:
            settings = collections.OrderedDict()

        # Read defaults.
        asta_dir = os.path.dirname(os.path.realpath(__file__))
        default_path = os.path.join(asta_dir, "defaults/astarc")
        assert os.path.isfile(default_path)
        defaultparser = configparser.ConfigParser()
        defaultparser = read_config(default_path)
        assert "MASTER" in defaultparser
        defaults = as_dict(defaultparser)["MASTER"]

        # Set defaults.
        for key, val in defaults.items():
            if key not in settings:
                settings[key] = val

        new_settings = collections.OrderedDict()
        for key, val in settings.items():
            new_val: Any = val
            if val in ("yes", "no"):
                # pylint: disable=simplifiable-if-expression
                new_val = True if val == "yes" else False
            new_settings[key.replace("-", "_")] = new_val
        settings = new_settings

        ox = Oxentiel(settings, mutable=True)
        _internal.ox = ox
    return ox



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Top-level typing classes. """
from asta import warning
from asta.array import Array
from asta.scalar import Scalar
from asta.decorators import typechecked
from asta.switchboard import Tensor, TFTensor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" A module for programmatically storing dimension sizes for annotations. """
import sys
from typing import Dict, Any, Union, Tuple
from asta.placeholder import Placeholder

# pylint: disable=redefined-outer-name, too-few-public-methods
# pylint: disable=no-self-use, too-many-ancestors


def __getattr__(name: str) -> Any:
    """ This exists solely to trick pylint. """
    raise NotImplementedError


class Shapes:
    """ An instance of this object acts as a proxy for this module. """

    def __init__(self) -> None:
        # Set any attributes here - before initialisation (they remain normal attrs).
        self.placeholder_map: Dict[str, Union[Placeholder, Tuple[int, ...]]] = {}

        # After initialization, setting attributes is the same as setting an item.
        self.__initialized = True

    def __getattr__(self, name: str) -> Union[Placeholder, Tuple[int, ...]]:
        if name in self.placeholder_map:
            return self.placeholder_map[name]
        placeholder = Placeholder(name)
        self.placeholder_map[name] = placeholder
        return placeholder

    def __setattr__(self, name: str, value: Any) -> None:
        """ Maps attributes to values. Only if we are initialised. """
        # This test allows attributes to be set in the ``__init__()`` method.
        if "_Shapes__initialized" not in self.__dict__:
            super().__setattr__(name, value)
        else:
            if not isinstance(value, tuple):
                raise TypeError("Value of a shape must be an tuple.")
            for element in value:
                if not isinstance(element, int):
                    raise TypeError("Shape elements must be integers.")
            self.placeholder_map[name] = value


sys.modules[__name__] = Shapes()  # type: ignore[assignment]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Prints a warning if typechecking will be silent. """
import os
from oxentiel import Oxentiel
from asta.config import get_ox
from asta.constants import Color

# pylint: disable=invalid-name

ox: Oxentiel = get_ox()
if "ASTA_TYPECHECK" in os.environ:
    ox.on = ox.on and os.environ["ASTA_TYPECHECK"] == "1"
if ox.on and not ox.print_passes:
    BORDER = "#" * 100
    BORDER = f"{Color.BOLD}{BORDER}{Color.END}"
    INNER = "-" * 100
    FILLER = "." * 100
    WARN = "ASTA TYPECHECKING IS ENABLED AND MAY CAUSE PERFORMANCE DEGRADATION"
    COLORED_WARN = f" {Color.RED}{WARN}{Color.END} "
    min_pad_size = 10
    pad_size = 100 - (len(WARN) + 2)
    side_size = max(pad_size // 2, min_pad_size)
    pad_parity = pad_size % 2 if side_size > 10 else 0
    left_padding = "=" * side_size
    right_padding = "=" * (side_size + pad_parity)
    PADDED_WARN = f"{left_padding}{COLORED_WARN}{right_padding}"

    print(BORDER)
    for _ in range(10):
        print(FILLER)
    print(INNER)
    print(PADDED_WARN)
    print(INNER)
    for _ in range(10):
        print(FILLER)
    print(BORDER)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Functions for checking type annotations and their origin types. """
from typing import Any, Tuple, Set, List, Union

from sympy.core.expr import Expr
from sympy.core.symbol import Symbol
from sympy.core.numbers import Number, Integer

from oxentiel import Oxentiel

import asta.dims
import asta.shapes
from asta.display import fail_uninitialized, fail_numerical_expression
from asta.constants import ALL_DIM_TYPES
from asta.placeholder import Placeholder

# pylint: disable=too-many-nested-blocks


def validate_placeholder_contents(
    contents: List[Union[Placeholder, Tuple[int, ...]]]
) -> None:
    """ Make sure all subplaceholders are not composite. """
    for elem in contents:
        # If it's a placeholder, it ought to be non-composite.
        if isinstance(elem, Placeholder):
            assert elem.name is not None
            assert len(elem.contents) == 0


def substitute(
    shape: Tuple[Any, ...], ox: Oxentiel
) -> Tuple[List[Any], Set[str], bool]:
    """ Parse a tuple shape, substituting placeholders for values. """
    dimension_sizes: List[Any] = []
    uninitialized_names: Set[str] = set()
    initialized = True

    for i, item in enumerate(shape):

        # Case 1: ``item`` is a sympy type.
        if isinstance(item, (Symbol, Expr)):
            expression = item

            # Use sympy to get a set of symbols used in expression.
            for symbol in item.free_symbols:

                # Check if any of the symbols in our list are in
                # ``asta.dims.symbol_map``.
                if symbol in asta.dims.symbol_map:
                    value = asta.dims.symbol_map[symbol]

                    # Out of those that are, we check if any have ``None``
                    # for their value.
                    if value is None:

                        # If so, we treat as before and raise an error.
                        name = symbol.name

                        # Prevents us from printing the same error message twice.
                        if name not in uninitialized_names:
                            fail_uninitialized(name, ox)
                        uninitialized_names.add(name)
                    else:
                        # Otherwise, we substitute in their values with sympy.
                        expression = expression.subs(symbol, value)

            # If this is a number (contains no symbols), it ought to be an integer.
            if isinstance(expression, Number):
                if not isinstance(expression, Integer):
                    fail_numerical_expression(shape[i], expression, ox)
                expression = int(expression)
            dimension_sizes.append(expression)

        # Case 2: ``item`` is a placeholder.
        elif isinstance(item, Placeholder):
            placeholder = item

            # What is item in case ``(*<placeholder>,) + (1,)``?
            # It will look like ``Array[<placeholder>, 1]``.
            if placeholder.name is not None:
                replacement: Union[Placeholder, Tuple[int, ...]]
                replacement = getattr(asta.shapes, placeholder.name)

                # Catch uninitialized placeholders.
                if isinstance(replacement, Placeholder):
                    initialized = False
                    name = placeholder.name
                    if name not in uninitialized_names:
                        fail_uninitialized(name, ox)
                    uninitialized_names.add(name)

                # Handle case where placeholder is unpacked in annotation.
                if placeholder.unpacked:
                    for elem in replacement:  # type: ignore[attr-defined]
                        dimension_sizes.append(elem)
                else:
                    dimension_sizes.append(replacement)
            else:
                # Treat case where placeholder has ``len(self.contents) > 1``.
                validate_placeholder_contents(placeholder.contents)
                contents = tuple(placeholder.contents)

                # Recursively call ``substitute()`` on the contents.
                sub_sizes, sub_uninit_names, sub_init = substitute(contents, ox)

                # Add in the results from recursive call.
                dimension_sizes.extend(sub_sizes)
                uninitialized_names = uninitialized_names.union(sub_uninit_names)
                initialized = initialized and sub_init

        # Case 3: ``item`` is a tuple.
        elif isinstance(item, tuple):

            # Recursively call ``substitute()`` on the tuple.
            sub_sizes, sub_uninit_names, sub_init = substitute(item, ox)

            # Add in the results from recursive call.
            dimension_sizes.extend(sub_sizes)
            uninitialized_names = uninitialized_names.union(sub_uninit_names)
            initialized = initialized and sub_init

        # Case 4: ``item`` is any other valid dimension type.
        elif isinstance(item, tuple(ALL_DIM_TYPES)):
            dimension_sizes.append(item)

        # Case 5: ``item`` has an invalid type.
        else:
            raise TypeError(f"Unsupported shape element type: '{type(item)}'")

    return dimension_sizes, uninitialized_names, initialized



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" Functions for checking type annotations and their origin types. """
import inspect
import collections
from io import TextIOBase, RawIOBase, IOBase, BufferedIOBase
from typing import (
    Any,
    Tuple,
    Set,
    Dict,
    List,
    AbstractSet,
    Callable,
    Union,
    IO,
    Sequence,
    Type,
    TextIO,
    BinaryIO,
)

from sympy.core.expr import Expr

from oxentiel import Oxentiel

from asta.utils import shapecheck, attrcheck
from asta.array import Array
from asta._array import _ArrayMeta
from asta.classes import SubscriptableMeta
from asta.display import (
    get_type_name,
    qualified_name,
    type_representation,
    pass_argument,
    fail_argument,
    fail_tuple,
    fail_tuple_length,
    fail_namedtuple,
    fail_empty_tuple,
    fail_list,
    fail_sequence,
    fail_dict,
    fail_set,
    fail_keys,
    fail_callable,
    fail_union,
    fail_class,
    fail_subclass,
    fail_io,
    fail_text_io,
    fail_binary_io,
    fail_complex,
    fail_float,
    fail_protocol,
    fail_literal,
    fail_too_many_args,
    fail_too_few_args,
    fail_fallback,
)
from asta.constants import torch, tf, _TORCH_IMPORTED, _TENSORFLOW_IMPORTED
from asta.substitution import substitute

METAMAP: Dict[type, SubscriptableMeta] = {_ArrayMeta: Array}

if _TORCH_IMPORTED:
    from asta.tensor import Tensor
    from asta._tensor import _TensorMeta

    METAMAP[_TensorMeta] = Tensor

if _TENSORFLOW_IMPORTED:
    from asta.tftensor import TFTensor
    from asta._tftensor import _TFTensorMeta

    METAMAP[_TFTensorMeta] = TFTensor

try:
    from typing import Literal  # type: ignore[attr-defined]
except ImportError:
    Literal = None

# pylint: disable=too-many-lines, too-many-nested-blocks, too-many-branches


def refresh(
    annotation: SubscriptableMeta, ox: Oxentiel
) -> Tuple[SubscriptableMeta, bool]:
    """ Load an asta type annotation containing classical placeholders. """
    dtype = annotation.dtype
    shape = annotation.shape
    dimvars: List[Any] = []
    initialized = True

    if annotation.shape is not None:

        dimvars, _, initialized = substitute(shape, ox)

        # Unpack tuple elements of ``dimvars``.
        unpacked: List[Any] = []
        for elem in dimvars:
            if isinstance(elem, tuple):
                unpacked = unpacked + list(elem)
            else:
                unpacked = unpacked + [elem]
        dimvars = unpacked

        shape = tuple(dimvars)
    # Note we're guaranteed that ``annotation`` has type ``SubscriptableMeta``.
    subscriptable_class = METAMAP[type(annotation)]

    refreshed_annotation: SubscriptableMeta
    if shape is not None:
        if dtype is not None:
            refreshed_annotation = subscriptable_class[dtype, shape]  # type: ignore
        else:
            refreshed_annotation = subscriptable_class[shape]  # type: ignore
    elif dtype is not None:
        refreshed_annotation = subscriptable_class[dtype]
    else:
        refreshed_annotation = annotation

    return refreshed_annotation, initialized


def check_asta(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check asta subscriptable class types. """

    annotation, initialized = refresh(annotation, ox)

    # If we haven't set the value of a placeholder used in the annotation, e.g.
    # we have ``Tensor[X, X + 1]`` but haven't yet set ``dims.X = 8``, then we
    # don't proceed with the check.
    if not initialized:
        return equations

    # Check if the literal ``value`` matches the annotation.
    rep: str = type_representation(value)

    # If the isinstance check fails, print/raise an error.
    if not isinstance(value, annotation):
        fail_argument(name, annotation, rep, ox)

    # Otherwise, print a pass.
    else:
        pass_argument(name, annotation, rep, ox)

        # Update equation set.
        shape_equations: Set[Expr] = set()
        attr_equations: Set[Expr] = set()

        if annotation.shape is not None:

            # HARDCODE
            # Handle case where type(shape) != tuple, e.g. ``torch.Size``.
            value_shape = tuple(value.shape)
            assert not isinstance(value_shape, (torch.Size, tf.TensorShape))

            # Grab equations from shapecheck call.
            shape_match, shape_equations = shapecheck(value_shape, annotation.shape)
            attr_match, attr_equations = attrcheck(value, annotation.kwattrs)
            assert shape_match and attr_match

        equations = equations.union(shape_equations, attr_equations)

    return equations


def check_typed_dict(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Typecheck a typed dict. """

    # If the argument is not a dict, we're aleady in trouble.
    if not isinstance(value, dict):
        fail_dict(name, annotation, qualified_name(value), ox)
        return equations

    expected_keys = frozenset(annotation.__annotations__)
    existing_keys = frozenset(value)

    extra_keys = existing_keys - expected_keys
    if extra_keys:
        fail_keys(name, expected_keys, existing_keys, ox)

    if annotation.__total__:
        missing_keys = expected_keys - existing_keys
        if missing_keys:
            fail_keys(name, expected_keys, existing_keys, ox)

    # pylint: disable=invalid-name
    MISSING_VALUE = object()
    for key, argtype in annotation.__annotations__.items():
        argvalue = value.get(key, MISSING_VALUE)
        if argvalue is not MISSING_VALUE:
            equations = check_annotation(
                f"{name}[{key}]", argvalue, argtype, equations, ox
            )
    return equations


def check_callable(name: str, value: Any, annotation: Any, ox: Oxentiel) -> None:
    """ Check an argument with annotation ``callable`` or ``Callable[]``. """
    if not callable(value):
        fail_callable(name, annotation, qualified_name(value), ox)

    if annotation.__args__:
        try:
            signature = inspect.signature(value)
        except (TypeError, ValueError):
            return

        if hasattr(annotation, "__result__"):
            # Python 3.5
            argument_types = annotation.__args__
            check_args = argument_types is not Ellipsis
        else:
            # Python 3.6+
            argument_types = annotation.__args__[:-1]
            check_args = argument_types != (Ellipsis,)

        if check_args:
            # The callable must not have keyword-only arguments without defaults.
            unfulfilled_kwonlyargs = [
                param.name
                for param in signature.parameters.values()
                if param.kind == inspect.Parameter.KEYWORD_ONLY
                and param.default == inspect.Parameter.empty
            ]
            if unfulfilled_kwonlyargs:
                # NOTE: Should this output a warning?
                return

            num_mandatory_args = len(
                [
                    param.name
                    for param in signature.parameters.values()
                    if param.kind
                    in (
                        inspect.Parameter.POSITIONAL_ONLY,
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    )
                    and param.default is inspect.Parameter.empty
                ]
            )
            has_varargs = any(
                param
                for param in signature.parameters.values()
                if param.kind == inspect.Parameter.VAR_POSITIONAL
            )

            # NOTE: Consider adding true recursive checking.
            if num_mandatory_args > len(argument_types):
                fail_too_many_args(name, len(argument_types), num_mandatory_args, ox)
            elif not has_varargs and num_mandatory_args < len(argument_types):
                fail_too_few_args(name, len(argument_types), num_mandatory_args, ox)
    return


def check_tuple(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check an argument with annotation ``tuple`` or ``Tuple[]``. """

    # Specialized check for NamedTuples.
    if hasattr(annotation, "_field_types"):

        if not isinstance(value, annotation):
            fail_namedtuple(name, qualified_name(annotation), qualified_name(value), ox)
            return equations

        # pylint: disable=protected-access
        for subname, field_type in annotation._field_types.items():
            equations = check_annotation(
                f"{name}.{subname}", getattr(value, subname), field_type, equations, ox
            )

    elif not isinstance(value, tuple):
        fail_tuple(name, qualified_name(value), ox)

    elif not getattr(annotation, "__args__", None):
        # Unparametrized Tuple or plain tuple.
        pass

    else:
        # Python 3.6+
        use_ellipsis = annotation.__args__[-1] is Ellipsis

        # Remove the ``Ellipsis`` from the end if it exists.
        if use_ellipsis:
            tuple_params = annotation.__args__[:-1]
        else:
            tuple_params = annotation.__args__

        if use_ellipsis:
            elem_type = tuple_params[0]
            for i, elem in enumerate(value):
                equations = check_annotation(
                    f"{name}[{i}]", elem, elem_type, equations, ox
                )
        elif tuple_params == ((),):
            if value != ():
                fail_empty_tuple(name, qualified_name(value), ox)
        else:
            if len(value) != len(tuple_params):
                ann_rep = qualified_name(annotation)
                fail_tuple_length(name, len(tuple_params), ann_rep, len(value), ox)
                return equations

            for i, (elem, elem_type) in enumerate(zip(value, tuple_params)):
                equations = check_annotation(
                    f"{name}[{i}]", elem, elem_type, equations, ox
                )

    return equations


def check_list(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check an argument with annotation ``list`` or ``List[]``. """
    if not isinstance(value, list):
        fail_list(name, annotation, qualified_name(value), ox)

    # If annotation is a subscriptable generic (``List[...]``).
    if annotation is not list:
        if annotation.__args__ not in (None, annotation.__parameters__):
            value_type = annotation.__args__[0]
            if value_type is not Any:

                value_iterable = value
                if not ox.check_all_sequence_elements:
                    value_iterable = value_iterable[:1]
                for i, v in enumerate(value_iterable):
                    element_equations = check_annotation(
                        f"{name}[{i}]", v, value_type, equations, ox
                    )
                    equations = equations.union(element_equations)

    return equations


def check_sequence(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check an argument with annotation ``Sequence[*]``. """
    if not isinstance(value, collections.abc.Sequence):
        fail_sequence(name, annotation, qualified_name(value), ox)

    # Consider removing this test or even just figuring out what it does?
    if annotation.__args__ not in (None, annotation.__parameters__):
        value_type = annotation.__args__[0]
        if value_type is not Any:

            value_iterable = value
            if not ox.check_all_sequence_elements:
                value_iterable = value_iterable[:1]
            for i, v in enumerate(value_iterable):
                element_equations = check_annotation(
                    f"{name}[{i}]", v, value_type, equations, ox
                )
                equations = equations.union(element_equations)

    return equations


def check_dict(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check an argument with annotation ``Dict[*]``. """
    if not isinstance(value, dict):
        fail_dict(name, annotation, qualified_name(value), ox)

    # Equivalently: if annotation is ``Dict[*]``.
    elif annotation is not dict:
        if annotation.__args__ not in (None, annotation.__parameters__):
            key_type, value_type = annotation.__args__
            if key_type is not Any or value_type is not Any:
                for k, v in value.items():
                    key_equations = check_annotation(
                        f"{name}.<key>", k, key_type, equations, ox
                    )
                    val_equations = check_annotation(
                        f"{name}[{k}]", v, value_type, equations, ox
                    )

                    equations = equations.union(key_equations, val_equations)

    return equations


def check_set(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check an argument with annotation ``Set[*]``. """
    if not isinstance(value, AbstractSet):
        fail_set(name, annotation, qualified_name(value), ox)

    # Equivalently: if annotation is ``Set[*]``.
    if annotation is not set:
        if annotation.__args__ not in (None, annotation.__parameters__):
            value_type = annotation.__args__[0]
            if value_type is not Any:
                for v in value:
                    set_equations = check_annotation(
                        f"{name}.<set_element>", v, value_type, equations, ox
                    )
                    equations = equations.union(set_equations)

    return equations


def check_union(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Typecheck an argument annotated with ``Union[]``. """
    if hasattr(annotation, "__union_params__"):
        # Python 3.5
        union_params = annotation.__union_params__
    else:
        # Python 3.6+
        union_params = annotation.__args__

    for type_ in union_params:
        try:
            equations = check_annotation(name, value, type_, equations, ox)
            return equations
        except TypeError:
            pass

    typelist = ", ".join(get_type_name(t) for t in union_params)
    fail_union(name, typelist, qualified_name(value), ox)
    return equations


def check_class(name: str, value: Any, annotation: Any, ox: Oxentiel) -> None:
    """ Check an arbitrary class. """
    if not inspect.isclass(value):
        fail_class(name, annotation, qualified_name(value), ox)

    # Needed on Python 3.7+
    if annotation is Type:
        return

    expected_class = annotation.__args__[0] if annotation.__args__ else None
    if expected_class and not issubclass(value, expected_class):
        supercls = qualified_name(expected_class)
        fail_subclass(name, supercls, qualified_name(value), ox)

    return


def check_number(name: str, value: Any, annotation: Any, ox: Oxentiel) -> None:
    """ Check for when ``value`` is in ``(complex, float, int)``. """
    if annotation is complex and not isinstance(value, complex):
        fail_complex(name, qualified_name(value.__class__), ox)
    elif annotation is float and not isinstance(value, float):
        fail_float(name, qualified_name(value.__class__), ox)


def check_io(name: str, value: Any, annotation: Any, ox: Oxentiel) -> None:
    """ Typecheck arguments with *IO annotations. """
    if annotation is TextIO:
        if not isinstance(value, TextIOBase):
            fail_text_io(name, qualified_name(value.__class__), ox)
    elif annotation is BinaryIO:
        if not isinstance(value, (RawIOBase, BufferedIOBase)):
            fail_binary_io(name, qualified_name(value.__class__), ox)
    elif not isinstance(value, IOBase):
        fail_io(name, qualified_name(value.__class__), ox)


def check_protocol(name: str, value: Any, annotation: Any, ox: Oxentiel) -> None:
    """ Typecheck arguments with protocol annotations. """
    if not issubclass(type(value), annotation):
        fail_protocol(name, type(value).__qualname__, annotation.__qualname__, ox)


def check_literal(name: str, value: Any, annotation: Any, ox: Oxentiel) -> None:
    """ Typecheck a value annotated with ``Literal[*]``. """
    if value not in annotation.__args__:
        fail_literal(name, annotation.__args__, value, ox)


# Equality checks are applied to these.
ORIGIN_TYPE_CHECKERS = {
    AbstractSet: check_set,
    dict: check_dict,
    Dict: check_dict,
    list: check_list,
    List: check_list,
    Sequence: check_sequence,
    collections.abc.Sequence: check_sequence,
    collections.abc.Set: check_set,
    set: check_set,
    Set: check_set,
    tuple: check_tuple,
    Tuple: check_tuple,
    Union: check_union,
}

# We ought to be above version 3.5.2.
assert Type is not None
NON_ASTA_INITIAL_CHECKERS = {
    Callable: check_callable,
    collections.abc.Callable: check_callable,
    type: check_class,
    Type: check_class,
}
if Literal is not None:
    NON_ASTA_CHECKERS = {**NON_ASTA_INITIAL_CHECKERS, **{Literal: check_literal}}
else:
    NON_ASTA_CHECKERS = NON_ASTA_INITIAL_CHECKERS


def check_annotation(
    name: str, value: Any, annotation: Any, equations: Set[Expr], ox: Oxentiel
) -> Set[Expr]:
    """ Check if ``value`` is of type ``annotation`` for asta types only. """

    # The solution to the set of equations for each individual annotation
    # should be unique. If there is no solution, print/raise an error. If there
    # are multiple solutions, raise an error; someone is trying to do inference
    # on way too many variables.

    # Thus there should be no need to iterate over the function signature more
    # than once. A single isinstance check should always be sufficient to
    # determine if there is a solution. However, that does not mean that single
    # isinstance checks are self-contained. In addition to specifying a unique
    # solution, all the annotations' solutions must agree for a given function
    # signature.

    # Get origin type.
    origin: Any = getattr(annotation, "__origin__", None)

    # Treat asta types.
    if isinstance(annotation, SubscriptableMeta):
        equations = check_asta(name, value, annotation, equations, ox)

    elif origin is not None:
        checker_function_map = ORIGIN_TYPE_CHECKERS
        if ox.check_non_asta_types:
            checker_function_map.update(NON_ASTA_CHECKERS)  # type: ignore
        checker_fn: Callable = checker_function_map[origin]
        if checker_fn:
            equations = checker_fn(name, value, annotation, equations, ox)
        else:
            equations = check_annotation(name, value, origin, equations, ox)
    elif inspect.isclass(annotation):

        subclass_callable: bool = issubclass(annotation, Callable)  # type: ignore
        has_args: bool = hasattr(annotation, "__args__")

        if issubclass(annotation, Tuple):  # type: ignore[arg-type]
            equations = check_tuple(name, value, annotation, equations, ox)
        elif issubclass(annotation, dict) and hasattr(annotation, "__annotations__"):
            equations = check_typed_dict(name, value, annotation, equations, ox)
        elif ox.check_non_asta_types:
            if issubclass(annotation, (float, complex)):
                check_number(name, value, annotation, ox)
            elif subclass_callable and has_args:
                # Needed on Python 3.5.0 to 3.5.2
                check_callable(name, value, annotation, ox)
            elif issubclass(annotation, IO):
                check_io(name, value, annotation, ox)
            elif getattr(annotation, "_is_protocol", False):
                check_protocol(name, value, annotation, ox)
            else:
                annotation = (
                    getattr(annotation, "__extra__", None) or origin or annotation
                )

                # As per https://github.com/python/typing/issues/552
                if annotation is bytes:
                    annotation = (bytearray, bytes)

                # Handles ``str`` and ``int``.
                if not isinstance(value, annotation):
                    fail_fallback(
                        name, qualified_name(annotation), qualified_name(value), ox
                    )

    return equations
